import sys
from functools import wraps
from pathlib import Path
from shutil import copyfile
from types import SimpleNamespace
from typing import Callable, Iterable, Optional

import click
import tomli

from vurf.parser import parse

APP_NAME = "VURF"
SECTION_ENV = f"{APP_NAME}_SECTION"
CONFIG_NAME = "config.toml"
PACKAGES_NAME = "packages.vurf"


section_option = click.option(
    "-s",
    "--section",
    required=False,
    envvar=SECTION_ENV,
    help=f"Defaults to `default_section` from config. Reads {SECTION_ENV} env variable.",
)


def ensure_config(quiet):
    config_path = Path(click.get_app_dir(APP_NAME))
    if not config_path.exists():
        config_path.mkdir(parents=True)
    config_file = config_path / CONFIG_NAME
    default_packages_file = Path.home() / PACKAGES_NAME
    if not config_file.is_file():
        if not quiet:
            click.secho("Config file not found...", fg="bright_black")
            click.secho(f"Creating default config {config_file}", fg="bright_black")
        copyfile(Path(__file__).parent / "config.toml", config_file)
    with config_file.open("rb") as opened:
        config = tomli.load(opened)
    if not Path(config["packages_location"]).is_file():
        if not quiet:
            click.secho("Packages file not found...", fg="bright_black")
            click.secho(f"Creating default {default_packages_file}", fg="bright_black")
            click.secho("You can change it later in the config.", fg="bright_black")
        copyfile(Path(__file__).parent / "default.vurf", default_packages_file)
    return config


def write_packages(ctx):
    with open(ctx.obj.config["packages_location"], "w") as f:
        f.write(ctx.obj.root.to_string())


def no_traceback(f: Callable) -> Callable:
    @wraps(f)
    def wrapper(*args, **kwds):
        try:
            return f(*args, **kwds)
        except Exception as e:
            sys.stderr.write(" ".join(e.args) + "\n")
            sys.exit(1)

    return wrapper


@click.group()
@click.pass_context
@click.option("-q", "--quiet", is_flag=True, help="Don't produce unnecessary output.")
@click.version_option(package_name=APP_NAME.lower())
@no_traceback
def main(ctx, quiet):
    config = ensure_config(quiet)
    with open(Path(config["packages_location"])) as f:
        root = parse(f)
    ctx.obj = ctx.ensure_object(SimpleNamespace)
    ctx.obj.config = config
    ctx.obj.root = root
    ctx.obj.quiet = quiet


@main.command(help="Add package(s).")
@section_option
@click.argument("packages", nargs=-1)
@click.pass_context
@no_traceback
def add(ctx, section: Optional[str], packages: Iterable[str]):
    if section is None:
        section = ctx.obj.config["default_section"]
    for package in packages:
        ctx.obj.root.add_package(section, package)
    write_packages(ctx)


@main.command(help="Remove package(s).")
@section_option
@click.argument("packages", nargs=-1)
@click.pass_context
@no_traceback
def remove(ctx, section: Optional[str], packages: Iterable[str]):
    if section is None:
        section = ctx.obj.config["default_section"]
    for package in packages:
        ctx.obj.root.remove_package(section, package)
    write_packages(ctx)


@main.command(help="Print default section.")
@click.pass_context
@no_traceback
def default(ctx):
    click.echo(ctx.obj.config["default_section"])


@main.command(help="Exit with indication if package is in packages.")
@section_option
@click.argument("package")
@click.pass_context
@no_traceback
def has(ctx, section: Optional[str], package: str):
    if section is None:
        section = ctx.obj.config["default_section"]
    ctx.exit(int(not ctx.obj.root.has_package(section, package)))


@main.command(help="Print list of packages.")
@section_option
@click.option(
    "--separator",
    required=False,
    type=str,
    default="\n",
    help='Section separator. Defaults to "\\n".',
)
@click.pass_context
@no_traceback
def packages(ctx, section, separator):
    click.echo(ctx.obj.root.get_packages(section, ctx.obj.config["parameters"], separator))


@main.command(help="Print list of sections.")
@click.option(
    "--separator",
    required=False,
    type=str,
    default="\n",
    help='Section separator. Defaults to "\\n".',
)
@click.pass_context
@no_traceback
def sections(ctx, separator):
    click.echo(ctx.obj.root.get_sections(separator))


@main.command(help="Install packages.")
@section_option
@click.pass_context
@no_traceback
def install(ctx, section):
    click.echo(
        ctx.obj.root.install(section, ctx.obj.config["sections"], ctx.obj.config["parameters"])
    )


@main.command(name="print", help="Print contents of packages file.")
@click.pass_context
@no_traceback
def print_(ctx):
    click.echo_via_pager(ctx.obj.root.to_string())


@main.command(help="Format packages file.")
@click.pass_context
@no_traceback
def format(ctx):
    write_packages(ctx)


@main.command(help="Edit packages file.")
@click.pass_context
@no_traceback
def edit(ctx):
    click.edit(filename=ctx.obj.config["packages_location"])


@main.command(help="Edit config file.")
@no_traceback
def config():
    click.edit(filename=f"{click.get_app_dir(APP_NAME)}/{CONFIG_NAME}")


if __name__ == "__main__":
    main()



import sys
from io import TextIOWrapper
from typing import cast

from vurf.nodes import Root
from vurf.parser.indenter import PythonesqueIndenter
from vurf.parser.stand_alone import Lark_StandAlone, ParseError
from vurf.parser.transformer import Root, VurfTransformer


def parse(file: TextIOWrapper) -> Root:
    parser = Lark_StandAlone(
        postlex=PythonesqueIndenter(),
        transformer=VurfTransformer(),
    )
    try:
        # Add extra newline in case there is none
        return cast(Root, parser.parse(file.read() + "\n"))
    except ParseError as e:
        sys.stderr.write(" ".join(e.args) + "\n")
        sys.exit(1)



from vurf.nodes import *
from vurf.parser.stand_alone import Transformer

__all__ = ["VurfTransformer"]


class VurfTransformer(Transformer):
    def comment_stmt(self, data):
        return Comment.from_parsed(data)

    def package_stmt(self, data):
        return Package.from_parsed(data)

    def ellipsis_stmt(self, data):
        return Ellipsis_.from_parsed(data)

    def if_stmt(self, data):
        return If.from_parsed(data)

    def elif_stmt(self, data):
        return Elif.from_parsed(data)

    def else_stmt(self, data):
        return Else.from_parsed(data)

    def with_stmt(self, data):
        return With.from_parsed(data)

    def file_input(self, data):
        return Root.from_parsed(data)



from vurf.parser.stand_alone import Indenter


class PythonesqueIndenter(Indenter):
    NL_type = "_NEWLINE"
    OPEN_PAREN_types = []
    CLOSE_PAREN_types = []
    INDENT_type = "_INDENT"
    DEDENT_type = "_DEDENT"
    tab_len = 2



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vurf', 'vurf.parser']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.0,<8.1.0', 'tomli>=2.0.0,<2.1.0']

entry_points = \
{'console_scripts': ['vurf = vurf.cli:main']}

setup_kwargs = {
    'name': 'vurf',
    'version': '0.10.0a3',
    'description': "Viliam's Universal Requirements Format",
    'long_description': '# VURF\n\n> Viliam\'s Universal Requirements Format\n\n## What it is\n*VURF* is a format, parser, and CLI for saving packages into Python-ish looking file.\n\n### Example packages.vurf\n```python\nwith pip:\n  vurf\n  black\n  if at_work:\n    ql-cq\n    ql-orange\nwith brew:\n  nnn  # terminal file manager\n```\n\n### Usage\n!TODO\n\n## Grammar\n*VURF* has [grammar](./vurf/grammar.lark) and LALR(1) parser implemented in [Lark](https://github.com/lark-parser/lark).\nIt aims to look like Python code as much as possible.\n\n### Keywords\n* `with [section]` - specifies "section" of requirements file. Different sections usually have different installers.\n* `if [condition]:` - conditions for including packages. See [Conditions](##Conditions) sections.\n* `elif [condition]:`\n* `else:`\n* `...` - ellipsis - placeholder for empty section.\n\n### Packages\n* are saved as `[name]  # [comment]`\n* `name` can be almost any valid package name (cannot start with "." or contain tabs or newline characters)\n* names containing spaces must be quoted. E.g. `\'multi word package name\'`\n* comments are optional\n\n## Config\n\nNote: *VURF* will automatically create config file on the first run.\n\n### Config file format\n```toml\n# Where packages file is saved\npackages_location = "/Users/viliam/packages.vurf"\n# Name of the default section\ndefault_section = "brew"\n\n# Sections can be though of as installers for different packages\n# Value of the section is the command for installing packages with `vurf install`\n[sections]\nbrew = "brew install"\ncask = "brew install --cask"\npython = "pip install --quiet --user"\n\n# Parameters are constants that can be accessed from conditionals\n[parameters]\nhostname = "mac"\nprimary_computer = true\nfs = "apfs"\n```\n\n## CLI\n```\n$ vurf\nUsage: vurf [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  -q, --quiet  Don\'t produce unnecessary output.\n  --version    Show the version and exit.\n  --help       Show this message and exit.\n\nCommands:\n  add       Add package(s).\n  config    Edit config file.\n  default   Print default section.\n  edit      Edit packages file.\n  format    Format packages file.\n  has       Exit with indication if package is in packages.\n  install   Install packages.\n  packages  Print list of packages.\n  print     Print contents of packages file.\n  remove    Remove package(s).\n  sections  Print list of sections.\n```\n\n## Conditions\n!TODO\n\n## Hooks\n!TODO\n',
    'author': 'Viliam Valent',
    'author_email': 'vurf@valent.email',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ViliamV/vurf',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
import pathlib
import subprocess
from functools import cached_property
from itertools import chain
from typing import Any, Optional, Union, cast


class Node:
    INDENT = "  "

    def __init__(self, data: str, children: list["Node"] = []) -> None:
        self.data = data
        self.children = children

    def __str__(self) -> str:
        return self.data

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.data == other.data

    def to_string(self, indent=0) -> str:
        indented = f"{self.INDENT * indent}{self}"
        return "\n".join(
            chain((indented,), (child.to_string(indent + 1) for child in self.children))
        )

    def add_child(self, pkg: "Node", *indexes: int) -> None:
        if not indexes:
            self.children.append(pkg)
        else:
            self.children[indexes[0]].add_child(pkg, *indexes[1:])

    def remove_child(self, node: "Node") -> None:
        self.children = [child for child in self.children if child != node]
        for child in self.children:
            child.remove_child(node)
            # Add ellipsis to empty withs and ifs
            if isinstance(child, (With, If, Elif, Else)) and not child.children:
                child.add_child(Ellipsis_("..."))

    def has_child(self, node: "Node") -> bool:
        if any(child == node for child in self.children):
            return True
        return any(child.has_child(node) for child in self.children)

    def get_packages(self, separator: str) -> str:
        return separator.join(
            child.package_name for child in self.children if isinstance(child, Package)
        ).strip(separator)


class Comment(Node):
    @classmethod
    def from_parsed(cls, data) -> "Comment":
        return cls(data[0].value)


class Package(Node):
    def __init__(self, data: str, comment: Optional["Comment"] = None) -> None:
        if data[0] == data[-1] and data[0] in {'"', "'"}:
            self._quoted = True
            data = data[1:-1]
        else:
            self._quoted = False
        super().__init__(data)
        self._comment = comment

    @classmethod
    def from_parsed(cls, data) -> "Package":
        return cls(data[0].value, data[1] if len(data) > 1 else None)

    @property
    def package_name(self) -> str:
        if self._quoted:
            return f"'{super().__str__()}'"
        return super().__str__()

    def __str__(self) -> str:
        if self._comment:
            return f"{self.package_name}{self.INDENT}{self._comment}"
        return self.package_name


class With(Node):
    @classmethod
    def from_parsed(cls, data) -> "With":
        arg, body = data
        return cls(arg.children[0].value, body.children)

    def __str__(self) -> str:
        return f"with {self.data}:"

    def get_packages(self, separator: str, parameters: dict[str, Any]) -> str:
        packages = [super().get_packages(separator)]
        for child in self.children:
            if isinstance(child, If):
                packages.append(child.get_packages(separator, parameters))
        return separator.join(packages).strip(separator)


class EvaluableMixin:
    def eval(self, parameters: dict[str, Any]) -> bool:
        return bool(eval(self.data, {"os": os, "pathlib": pathlib}, parameters))


class If(Node, EvaluableMixin):
    def __init__(
        self,
        data: str,
        children: list["Node"],
        branches: list[Union["Elif", "Else"]] = [],
    ) -> None:
        super().__init__(data, children=children)
        self.branches = branches

    def get_packages(self, separator: str, parameters: dict[str, Any]) -> str:
        self_is_true = self.eval(parameters)
        packages = [super().get_packages(separator)] if self_is_true else []
        for branch in self.branches:
            if isinstance(branch, Elif) and branch.eval(parameters):
                packages.append(branch.get_packages(separator))
            if isinstance(branch, Else) and not self_is_true:
                packages.append(branch.get_packages(separator))
        return separator.join(packages)

    @classmethod
    def from_parsed(cls, data) -> "If":
        arg, body = data[0], data[1]
        return cls(arg.children[0].value, body.children, data[2:])

    def __str__(self) -> str:
        return f"if {self.data}:"

    def to_string(self, indent=0) -> str:
        return "\n".join(
            chain(
                (super().to_string(indent=indent),),
                (branch.to_string(indent) for branch in self.branches),
            )
        )


class Elif(Node, EvaluableMixin):
    @classmethod
    def from_parsed(cls, data) -> "Elif":
        arg, body = data
        return cls(arg.children[0].value, body.children)

    def __str__(self) -> str:
        return f"elif {self.data}:"


class Else(Node):
    @classmethod
    def from_parsed(cls, data) -> "Else":
        return cls("else:", data[0].children)


class Ellipsis_(Node):
    @classmethod
    def from_parsed(cls, data) -> "Ellipsis_":
        return cls(data[0].value)


class Root:
    def __init__(self, children: list["Node"] = []) -> None:
        self.children = children

    @classmethod
    def from_parsed(cls, data) -> "Root":
        return cls(data)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}"

    def to_string(self, indent=0) -> str:
        return "\n".join(child.to_string(indent) for child in self.children) + "\n"

    @cached_property
    def section_indexes(self) -> dict[str, int]:
        return {child.data: i for i, child in enumerate(self.children) if isinstance(child, With)}

    def _get_section_by_name(self, section_name: str) -> "With":
        try:
            return cast(With, self.children[self.section_indexes[section_name]])
        except KeyError:
            raise KeyError(f'No section named "{section_name}"')

    def get_sections(self, separator: str) -> str:
        return separator.join(self.section_indexes.keys())

    def add_package(self, section_name: str, package_name: str, *indexes: int) -> None:
        section = self._get_section_by_name(section_name)
        args = package_name.split(maxsplit=1)
        comment = Comment(args[1].strip()) if len(args) > 1 else None
        package = Package(args[0].strip(), comment)
        # Ensure no duplicates
        if not section.has_child(package):
            section.add_child(package, *indexes)

    def has_package(self, section_name: str, package_name: str) -> bool:
        section = self._get_section_by_name(section_name)
        args = package_name.split(maxsplit=1)
        comment = Comment(args[1].strip()) if len(args) > 1 else None
        package = Package(args[0].strip(), comment)
        return section.has_child(package)

    def get_packages(
        self, section: Optional[str], parameters: dict[str, Any], separator: str = " "
    ) -> str:
        if section is not None:
            return self._get_section_by_name(section).get_packages(separator, parameters)
        return separator.join(
            self.get_packages(section, parameters, separator)
            for section in self.section_indexes.keys()
        )

    def install(
        self,
        section: Optional[str],
        section_commands: dict[str, str],
        parameters: dict[str, Any],
    ) -> None:
        if section is not None:
            packages = self._get_section_by_name(section).get_packages(" ", parameters)
            command = section_commands[section]
            subprocess.run(f"{command} {packages}", shell=True)
            return
        for section in self.section_indexes.keys():
            self.install(section, section_commands, parameters)

    def remove_package(self, section: str, package: str) -> None:
        self._get_section_by_name(section).remove_child(Package(package))
